require.paths.unshift(__dirname + '/../../dep');

// re-export connect middleware which was re-exported by express
var exports = module.exports = require('connect').middleware;

// dependencies
var Step = require('step'),
    fs = require('fs'),
    sys = require('sys'),
    express = require('express'),
    underscore = require('underscore'),
    underscoreStrings = require('underscore.strings'),
    couchdb = require('couchdb'),
    helpers = require('./helpers'),
    RedisStore = require('connect-redis'),
    Stores = require('./storage').Stores;

exports.createServer = function() {
    return new express.Server(Array.prototype.slice.call(arguments));
};

var boot = exports.boot = function(app, options, fn) {
    var cmdLineOptions = getOptions();
    var subAppSettings = require(options.homeDir + '/settings');
    var settings = _.extend(app.settings, options, subAppSettings, cmdLineOptions);
    app.settings = settings;

    var realConsoleInfo;
    if (!settings.verbose) {
        // keep track of console.info so we can give it back when we're done
        realConsoleInfo = console.info;
        // if verbose is turned off mute console.info()
        console.info = function() {};
    }
    
    console.info('Booting M1 Node');

    // create an app namespace for exporting common values
    var appNamespace = {};
    // alias for internal use
    global.m1node = appNamespace;
    // alias for app use
    if (settings.name) {
        global[settings.name] = appNamespace;
    }

    // export app settings
    appNamespace.settings = settings;
    // create storage and export
    var stores = new Stores(settings.databases, settings.databaseAuthFile);
    appNamespace.stores = stores;

    function flushDatabases(fn) {
        if (!settings.flush) {
            return fn();
        }
        console.info('Flushing databases');
        var flushDatabase = function(store, fn) {
            var name = store.getKey();
            console.info(' * Flushing database ' + name);
            var db = store.getDatabase();
            db.remove(function(err, ok) {
                db.create(function(err, ok) {
                    var conf = store.getConf();
                    db.request({
                        method: 'put',
                        path: '/_security',
                        data: {
                            admins: {
                                names: [conf.username],
                                roles: []
                            },
                            readers: {
                                names: [conf.username],
                                roles: []
                            }
                        }
                    }, fn);
                });
            });
        };
            
        Step(
            function start() {
                // prime the pump
                return null;
            },
            function flush(err, results) {
                if (err) { throw err; }
                var group = this.group();
                _.each(stores, function(store) {
                    flushDatabase(store, group());
                });
            },
            function end(err, vals) {
                if (err) { throw err; }
                fn();
            }
        );
    }

    function bootSubApps(fn) {
        console.info('Booting sub apps');
        var boot;
        try {
            boot = require(settings.homeDir + '/boot');
        } catch (err) {
            return fn();
        }
        boot(app, fn);
    }

    function bootApplication(fn) {
        console.info('Booting application');
        
        // logging middleware
        if (settings.env !== 'test') {
            app.use(express.logger({ format: ':method :url :status' }));
        }

        // plumbing middleware
        app.use(express.bodyDecoder());
        app.use(express.methodOverride());
        app.use(express.cookieDecoder());

        // session middleware
        var oneHour = 60 * 60 * 1000;
        app.use(express.session({
            key: settings.name.toLowerCase() + '.sid',
            store: settings.sessionStore || new RedisStore({
                maxAge: settings.sessionTimeout || oneHour
            })
        }));

        // routing middleware
        app.use(app.router);

        // serve static content
        app.use(express.staticProvider(settings.homeDir + '/public'));

        // deal with app errors
        function NotFound(path) {
            this.name = 'NotFound';
            var msg = 'Not Found';
            if (path) {
                msg = 'Cannot find ' + path;
                this.path = path;
            }
            Error.call(this, msg);
            Error.captureStackTrace(this, arguments.callee);
        }

        // handle error if it's a missing page or pass it on
        app.use(function(req, res, next) {
            next(new NotFound(req.url));
        });
        app.error(function(err, req, res, next) {
            if (err instanceof NotFound) {
                res.render('404', {
                    status: 404,
                    error: err
                });
            } else {
                next(err);
            }
        });

        // handle errors in general
        app.error(function(err, req, res) {
            console.dir(err);
            res.render('500', { status: 500, error: err });
        });

        // setup template engine
        var templateLanguage = settings.template_language || 'ejs';
        app.set('views', settings.homeDir + '/views');
        app.register('.html', require(templateLanguage));
        app.set('view engine', 'html');

        // setup framework level dynamic helpers
        app.dynamicHelpers({
            request: function(req) {
                return req;
            },
            messages: function(req, res) {
                return function() {
                    var messages = req.flash();
                    return res.partial('messages', {
                        object: messages,
                        as: 'messageTypes',
                        locals: { hasMessages: Object.keys(messages).length },
                        dynamicHelpers: false
                    });
                };
            }
        });
        fn();
    }

    function bootCouchChanges(fn) {
        console.info('Booting couch change monitor');
        var changes, db, stream;
        _.each(stores, function(store, name) {
            changes = null;
            try {
                changes = require(settings.homeDir + '/store/changes/' + name);
            } catch(err) {}
            if (changes) {
                db = store.getDatabase();
                _.each(changes, function(listener, name) {
                    stream = db.changesStream();
                    stream.addListener('data', listener);
                });
            }
        });
        fn();
    }

    function bootCouchViews(fn) {
        console.info('Booting couch views');
        
        // import views or give up
        try {
            var views = require(settings.homeDir + '/store/views');
        } catch(err) {
            return fn();
        }
        
        var bootCouchView = function(store, fn) {
            var name = store.getKey();
            var db = store.getDatabase();
            if (views[name]) {
                _.each(views[name], function(design, designName) {
                    var id = '_design/' + designName;
                    var saveDesignDoc = {
                        id: id, 
                        views: design
                    };
                    var noChange = false;
                    db.getDoc(id, function(err, designDoc) {
                        if (designDoc && couchdb.toJSON(designDoc.views) !== couchdb.toJSON(saveDesignDoc.views)) {
                            saveDesignDoc._rev = designDoc._rev;
                        } else if (!err) {
                            noChange = true;
                        }
                        if (noChange) {
                            fn(null, false);
                        } else {
                            db.saveDesign(id, saveDesignDoc, function(err, ok) {
                                if (err) { console.dir(err); }
                                fn(null, {
                                    name: name,
                                    id: id
                                });
                            });
                        }
                    });
                });
            } else {
                fn(null, false);
            }
        };
        
        Step(
            function start() {
                // prime the pump
                return null;
            },
            function views(err, results) {
                if (err) { throw err; }
                var group = this.group();
                _.each(stores, function(store) {
                    bootCouchView(store, group());
                });
            },
            function end(err, vals) {
                if (err) { throw err; }
                var updates = _.compact(vals);
                if (updates.length > 0) {
                    _.each(updates, function(update) {
                        console.info(' * Updated ' + update.id + ' in ' + update.name + ' database');
                    });
                } else {
                    console.info(' * All views are up to date');
                }
                fn();
            }
        );
    }

    function bootRoutes() {
        console.info('Booting routes');
        var urls = require(settings.homeDir + '/urls').patterns;
        _.each(urls, function(route) {
            bootRoute(route);
        });
    }

    function parseRoute(rawRoute) {
        var pattern = rawRoute[0];
        var rawAction = rawRoute[1];
        var extras = {};
        var middleware = [];
        var thirdArg = rawRoute[2];
        if (rawRoute.length === 4) {
            middleware = rawRoute[2];
            extras = rawRoute[3];
        } else if (_.isDefined(thirdArg)) {
            if (!_.isFunction(thirdArg) && !_.isArray(thirdArg)) {
                extras = thirdArg;
            } else {
                middleware = thirdArg;
            }
        }
        if (!_.isArray(middleware)) {
            middleware = [middleware];
        }
        var parts = rawAction.split('.');
        var actionName = parts[(parts.length-1)];
        var controllerParts = parts.slice(0, (parts.length-1));
        var controllerPath = controllerParts.join('/');
        var controller = require(settings.homeDir + '/controllers/' + controllerPath);
        var action = controller[actionName];
        return {
            pattern: pattern,
            action: action,
            controllerPath: controllerPath,
            actionName: actionName,
            middleware: middleware,
            extras: extras
        };
    }

    function registerRoute(route, fn) {
        app.get(route.pattern, route.middleware, fn);
        app.post(route.pattern, route.middleware, fn);
    }

    function addRouteMiddleware(route, fn) {
        route.middleware.push(function(req, res, next) {
            // TODO--isn't there some way to merge this with request.params?
            req.extras = route.extras;
            next();
        });

        var middlewares = require('./middleware/profile');
        _.each(middlewares, function(middleware) {
            route.middleware.push(middleware);
        });

        // add app middleware
        var appMiddleware;
        try {
            appMiddleware = require(settings.homeDir + '/middleware');
        } catch(err) {
            return;
        }
        _.each(appMiddleware, function(middleware) {
            route.middleware.push(middleware);
        });
    }

    function bootRoute(rawRoute) {
        var route = parseRoute(rawRoute),
            fn = controllerAction(route);
        addRouteMiddleware(route);
        registerRoute(route, fn);
    }

    function controllerAction(route) {
        return function(req, res, next) {
            var render = res.render,
                format = req.params.format,
                path = settings.homeDir + '/views/' + route.controllerPath + '/' + route.actionName + '.html';
            res.render = function(obj, options, fn) {
                res.render = render;
                // template path
                if (_.isString(obj)) {
                    return res.render(obj, options, fn);
                }
                // format support
                if (format) {
                    if (format === 'json') {
                        var contentType = { 'Content-Type': 'application/json; charset=utf8' };
                        return res.send(obj, contentType);
                    } else {
                        throw new Error('unsupported format "' + format + '"');
                    }
                }
                // render template
                res.render = render;
                options = options || {};
                options.locals = options.locals || {};
                options.locals[route.controllerPath] = obj;
                return res.render(path, options, fn);
            };
            route.action.apply(this, arguments);
        };
    }

    function bootDynamicHelpers() {
        var dynamicHelpers;
        try {
            dynamicHelpers = require(options.homeDir + '/dynamicHelpers');
        } catch(err) {
            return;
        }
        app.dynamicHelpers(dynamicHelpers);
    }

    Step(
        function prime() {
            return 1;
        },
        function dynamicHelpers(err, val) {
            if (err) { throw err; }
            bootDynamicHelpers();
            return true;
        },
        function flush(err, val) {
            if (err) { throw err; }
            flushDatabases(this);
        },
        function views(err, val) {
            if (err) { throw err; }
            bootCouchViews(this);
        },
        function changes(err, val) {
            if (err) { throw err; }
            bootCouchChanges(this);
        },
        function application(err, val) {
            if (err) { throw err; }
            bootApplication(this);
        },
        function routes(err, val) {
            if (err) { throw err; }
            bootRoutes();
            return true;
        },
        function subs(err, val) {
            if (err) { throw err; }
            bootSubApps(this);
        },
        function finale(err, val) {
            if (err) { throw err; }
            // put console info back where it belongs
            if (!settings.verbose) {
                console.info = realConsoleInfo;
            }
            fn();
        }
    );
};
