var path = require('path'),
    m1node = require('m1node'),
    underscore = require('underscore'),
    should = require('should'),
    express = require('express'),
    Step = require('step'),
    assert = require('assert'),
    MemoryStore = m1node.session.MemoryStore,
    assertLintFree = require('node-assert-lint-free');

var minimalOptions = {
    homeDir: path.join(__dirname, '/../example/minimal'),
    sessionStore: new MemoryStore({ reapInterval: -1 }),
    flush: true
};

var minimalOptionsNoFlush = {
    homeDir: path.join(__dirname, '/../example/minimal'),
    sessionStore: new MemoryStore({ reapInterval: -1 }),
    flush: false
};

var fullOptions = {
    homeDir: path.join(__dirname, '/../example/full'),
    sessionStore: new MemoryStore({ reapInterval: -1 }),
    flush: true
};

var getStores = function() {
    var settings = require('../example/minimal/settings');
    var stores = new Stores(settings.databases, settings.databaseAuthFile);
    return stores;
};

var addTestObjectsToDatabases = function(fn) {
    
    var testObj = {
        stamp: new Date().getTime(),
        test1: 'test 1'
    };
    
    Step(
        function start() {
            // prime the pump
            return null;
        },
        function addTestObjs(err, results) {
            if (err) { throw err; }
            var group = this.group();
            var stores = getStores();
            var storeKeys = Object.keys(stores);
            storeKeys.forEach(function(key) {
                var store = stores[key];
                var db = store.getDatabase();
                db.saveDoc('testObj', testObj, group());
            });
            
        },
        function end(err, vals) {
            if (err) { throw err; }
            fn(err, vals);
        }
    );
};

var getTestObjectsFromDatabases = function(fn) {
    Step(
        function start() {
            // prime the pump
            return null;
        },
        function getTestObjs(err, results) {
            if (err) { throw err; }
            var group = this.group();
            var stores = getStores();
            var storeKeys = Object.keys(stores);
            storeKeys.forEach(function(key) {
                var store = stores[key];
                var db = store.getDatabase();
                db.getDoc('testObj', group());
            });
        },
        function end(err, vals) {
            if (err && err.reason !== 'missing') { throw err; }
            fn(null, _.compact(vals));
        }
    );
};
    
var removeDatabases = function(fn) {
    Step(
        function start() {
            // prime the pump
            return null;
        },
        function deleteTestDbs(err, results) {
            if (err) { throw err; }
            var group = this.group();
            var stores = getStores();
            var storeKeys = Object.keys(stores);
            storeKeys.forEach(function(key) {
                var store = stores[key];
                var db = store.getDatabase();
                db.remove(group());
            });
        },
        function end(err, vals) {
            if (err && err.reason !== 'missing') { throw err; }
            fn(null, _.compact(vals));
        }
    );
};

var tests = {
    'test lint free': function() {
        var opts = {
            ignore: [
                "Avoid arguments.callee.",
                "Reserved name \'__proto__\'.",
                "Expected an identifier and instead saw 'instanceof' (a reserved word).",
                "Expected an identifier and instead saw 'arguments' (a reserved word)."
            ]
        };
        var roots = [
            path.join(__dirname),
            path.join(__dirname, '../lib'),
            path.join(__dirname, '../example')
        ];
        assertLintFree(roots, opts);
    },
    
    'test inheritance': function() {
        var server = m1node.createServer();
        server.should.be.an.instanceof(express.Server);
    },
    
    'test express middleware autoloaders': function() {
        m1node.errorHandler.should.equal(express.errorHandler);
    },
    
    'test minimal boot': function(done) {
        var app = m1node.createServer();
    
        m1node.boot(app, minimalOptions, function() {
            assert.ok(true);
            done();
        });
    },
    
    'test views minimal': function(done) {
        var app = m1node.createServer();
    
        m1node.boot(app, minimalOptions, function() {
            // root view
            assert.response(app, {
                url: '/', timeout: 500
            }, {
                status: 200
            }, function(res) {
                res.body.should.include.string('<title>m1node minimal example app</title>');
                res.body.should.include.string('<div>An example view</div>');
                done();
            });
        });
    },
    
    'test settings merge': function(done) {
        var app = m1node.createServer();
    
        minimalOptions.settingsMergeTestFromOptions = 'settings merge test from options';
    
        m1node.boot(app, minimalOptions, function() {
            // NOTE: I'd love to test that incoming command line params merge correctly but unsure how
            var settingsMergeTestFromOptions = app.set('settingsMergeTestFromOptions');
            var settingsMergeTestFromSettingsFile = app.set('settingsMergeTestFromSettingsFile');
    
            settingsMergeTestFromOptions.should.eql('settings merge test from options');
            settingsMergeTestFromSettingsFile.should.eql('settings merge test from settings file');
            done();
        });
    },
    
    'test flushing databases': function(done) {
        var app = m1node.createServer();
        m1node.boot(app, minimalOptions, function() {
            // add anything to the database
            addTestObjectsToDatabases(function() {
                getTestObjectsFromDatabases(function(err, testObjs) {
                    testObjs.should.have.length(2);
                    // now boot app which should flush everything
                    var app2 = m1node.createServer();
                    m1node.boot(app2, minimalOptions, function() {
                        // make sure there's nothing there
                        getTestObjectsFromDatabases(function(err, testObjs2) {
                            testObjs2.should.have.length(0);
                            done();
                        });
                    });
                });
            });
        });
    },
    
    'test minimal boot no flush': function(done) {
        var app = m1node.createServer();
        minimalOptions.flush = false;
        m1node.boot(app, minimalOptions, function() {
            // add anything to the database
            addTestObjectsToDatabases(function() {
                getTestObjectsFromDatabases(function(err, testObjs) {
                    testObjs.should.have.length(2);
                    // now boot app which should flush everything
                    var app2 = m1node.createServer();
                    m1node.boot(app2, minimalOptions, function() {
                        // make sure there's nothing there
                        getTestObjectsFromDatabases(function(err, testObjs2) {
                            minimalOptions.flush = true; // set it back
                            testObjs2.should.have.length(2);
                            done();
                        });
                    });
                });
            });
        });
    },
    
    'test flushing non-existent databases': function(done) {
        var app = m1node.createServer();
        m1node.boot(app, minimalOptions, function() {
            removeDatabases(function() {
                tests['test flushing databases'](done);
            });
        });
    },
    
    'test full boot': function(done) {
        var app = m1node.createServer();
    
        m1node.boot(app, fullOptions, function() {
            assert.ok(true);
            done();
        });
    },
    
    'test root view full': function(done) {
        var app = m1node.createServer();
    
        m1node.boot(app, fullOptions, function() {
    
            assert.response(app, {
                url: '/', timeout: 500
            }, {
                status: 200
            }, function(res) {
                res.body.should.include.string('<div>An example view</div>');
                done();
            });
        });
    },
    
    'test non-existent view full': function(done) {
        var app = m1node.createServer();
    
        m1node.boot(app, fullOptions, function() {
            assert.response(app, {
                url: '/nopage/', timeout: 500
            }, {
                status: 404
            }, function(res) {
                res.body.should.include.string('<h1>Cannot find /nopage/</h1>');
                done();
            });
            
        });
    },
    
    'test error view full': function(done) {
        var app = m1node.createServer();
    
        m1node.boot(app, fullOptions, function() {
            assert.response(app, {
                url: '/err/', timeout: 500
            }, {
                status: 500
            }, function(res) {
                res.body.should.include.string('<p>Sorry an error has occurred, please try refreshing the page or navigate back to <a href="/">home</a>.</p>');
                res.body.should.include.string('<p>Sorry an error has occurred, please try refreshing the page or navigate back to <a href="/">home</a>.</p>');
                done();
            });
        });
    },
    
    'test middleware loading': function(done) {
        var app = m1node.createServer();
        
        m1node.boot(app, fullOptions, function() {
            assert.response(app, {
                url: '/middlewaretest/', timeout: 6000
            }, {
                status: 200
            }, function(res) {
                res.body.should.include.string('<p><strong>appLevelMiddlewareTestValue</strong>: app level middleware test value</p>');
                done();
            });
        });
    },
    
    'test middleware and extras configuration': function(done) {
        var app = m1node.createServer();
        
        m1node.boot(app, fullOptions, function() {
            assert.response(app, {
                url: '/hasmiddlewareandextrasconfigtest/', timeout: 6000
            }, {
                status: 200
            }, function(res) {
                res.body.should.include.string('<p><strong>requestExtraVal</strong>: request extra test val</p>');
                res.body.should.include.string('<p><strong>urlPatternLevelMiddlewareTestValue</strong>: url pattern level middleware test value</p>');
                done();
            });
        });
    },
    
    'test request extras from url patterns': function(done) {
        var app = m1node.createServer();
        
        m1node.boot(app, fullOptions, function() {
            assert.response(app, {
                url: '/requestextrastest/', timeout: 500
            }, {
                status: 200
            }, function(res) {
                res.body.should.include.string('<p><strong>requestExtraVal</strong>: request extra test val</p>');
                done();
            });
        });
    },
    
    'test dynamic helpers loaded': function(done) {
        var app = m1node.createServer();
        m1node.boot(app, fullOptions, function() {
            assert.response(app, {
                url: '/dynamichelpertest/', timeout: 500
            }, {
                status: 200
            }, function(res) {
                res.body.should.include.string('<p><strong>testHelper</strong>: test helper value</p>');
                done();
            });
        });
    },
    
    'test couch views loaded': function(done) {
        // NOTE: there's a lot more than could be done here but for now we just check 
        // that the database has a view document to verify it got loaded
        var app = m1node.createServer();
        m1node.boot(app, fullOptions, function() {
            var stores = getStores();
            var db = stores.m1nodeTest1.getDatabase();
            db.getDoc('_design/test', function(err, doc) {
                assert.isDefined(doc.views.all.map, "'test' design document view 'all' should be loaded");
                done();
            });
        });
    },
    
    'test sub app boot': function(done) {
        var app = m1node.createServer();
        m1node.boot(app, fullOptions, function() {
            app.set('subAppBootValue').should.include.string('a value for testing sub app boot');
            done();
        });
    },
    
    'test session get and set': function(done) {
        var app = m1node.createServer();
        var cookie;
        var testUsername = 'testuser' + new Date().getTime();
        m1node.boot(app, fullOptions, function() {
            assert.response(app, {
                url: '/sessionsettest/?username=' + testUsername,
                timeout: 500
            },{
                status: 200
            }, function(res) {
                assert.response(app, {
                    url: '/sessiongettest/',
                    headers: {
                        "cookie": res.headers['set-cookie']
                    },
                    timeout: 500
                }, {
                    status: 200
                }, function(res) {
                    res.body.should.include.string('<p><strong>session.username</strong>: ' + testUsername + '</p>');
                    done();
                });
            });
            
        });
    },
    
    'test profile middleware': function(done) {
        var app = m1node.createServer();
        var cookie;
        var testUsername = 'testuser' + new Date().getTime();
        m1node.boot(app, fullOptions, function() {
            assert.response(app, {
                url: '/sessionsettest/?username=' + testUsername,
                timeout: 500
            },{
                status: 200
            }, function(res) {
                assert.response(app, {
                    url: '/profilemiddlewaresetvaluetest/',
                    headers: {
                        "cookie": res.headers['set-cookie']
                    },
                    timeout: 500
                }, {
                    status: 200
                }, function(res) {
                    assert.response(app, {
                        url: '/profilemiddlewaregetvaluetest/',
                        headers: {
                            "cookie": res.headers['set-cookie']
                        },
                        timeout: 500
                    }, {
                        status: 200
                    }, function(res) {
                        res.body.should.include.string('<p><strong>profileMiddlewareSetValue</strong>: profile middleware test value</p>');
                        done();
                    });
                });
            });
            
        });
    },
    
    'test no autoload middleware profile': function(done) {
        var app = m1node.createServer();
        var cookie;
        var testUsername = 'testuser' + new Date().getTime();
        m1node.boot(app, fullOptions, function() {
            assert.response(app, {
                url: '/sessionsettest/?username=' + testUsername,
                timeout: 500
            },{
                status: 200
            }, function(res) {
                assert.response(app, {
                    url: '/noautoloadprofiletest/',
                    headers: {
                        "cookie": res.headers['set-cookie']
                    },
                    timeout: 500
                }, {
                    status: 200
                }, function(res) {
                    res.body.should.include.string('<p><strong>profile</strong>: no profile present</p>');
                    done();
                });
            });
        });
    }
};

module.exports = tests;
